/**
 * This is a static class that manages the Air Kilometres program.
 * 
 * @author Alan Williams
 */
public class AirKM
{
	// CLASS VARIABLES

	/**
	 * Bonus factor for multiplying points during purchases.
	 * <p>
	 * Editorial comment: This value really ought to be private and have an
	 * accessor method, according to a style checking program.
	 */
	public static int POINTS_FACTOR = 1;

	/**
	 * Catalogue that keeps references to all products.
	 */
	private static Product[] catalogue;

	/**
	 * Journal that keeps references to all transactions.
	 */
	private static Transaction[] journal;

	/**
	 * Number of items currently in catalogue.
	 */
	private static int numCatalogueItems;

	/**
	 * Number of items currently in journal.
	 */
	private static int numJournalItems;

	/**
	 * Maximum size for journal and catalogue arrays.
	 */
	private static int maxSize;

	// CLASS METHODS:

	/**
	 * Initialize the Air Kilometres journal and catalogue arrays.
	 * <p>
	 * The maximum size of each array is set to the specified value.
	 * 
	 * @param arrayMaxSize
	 *            The maximum array size.
	 */
	public static void initialize( int arrayMaxSize )
	{
		AirKM.numCatalogueItems = 0;
		AirKM.numJournalItems = 0;
		AirKM.maxSize = arrayMaxSize;
		AirKM.catalogue = new Product[AirKM.maxSize];
		AirKM.journal = new Transaction[AirKM.maxSize];
	}

	/**
	 * Method to change the points factor.
	 * <p>
	 * A password is required to change the points factor. The new value should
	 * be one of 1, 2, or 3.
	 * 
	 * @param newFactor
	 *            The new points factor.
	 * @param password
	 *            Guess at the password to change the points factor.
	 */
	public static void setPointsFactor( int newFactor, String password )
	{
		// Check password before anything else.

		if ( password.compareTo( "ITI 1220 rules!" ) == 0 )
		{
			if ( ( newFactor >= 1 ) && ( newFactor <= 3 ) )
			{
				AirKM.POINTS_FACTOR = newFactor;
				System.out
						.println( "Confirmation:  points factor has been modified." );
			}
			else
			{
				System.out.println( "Invalid points factor." );
			}
		}
		else
		{
			System.out.println( "Error: Incorrect password." );
		}
	}

	/**
	 * Adds a new product to the catalogue.
	 * 
	 * @param newProduct
	 *            Product to be added.
	 * @return True if there was space in the catalogue for the new product, and
	 *         false if the catalogue is already full.
	 */
	public static boolean addProduct( Product newProduct )
	{
		boolean result;

		// Check if there is space left in the catalogue.

		if ( AirKM.numCatalogueItems < AirKM.maxSize )
		{
			catalogue[AirKM.numCatalogueItems] = newProduct;

			// Increment number of catalogue entries.

			AirKM.numCatalogueItems = AirKM.numCatalogueItems + 1;
			result = true;
		}
		else
		{
			System.out.println( "Error:  Product catalogue is full." );
			result = false;
		}
		return result;
	}

	/**
	 * Returns the product at the specified index in the catalogue.
	 * 
	 * @param number
	 *            Index of requested product.
	 * @return The product that is requested, or null if there is no such
	 *         product.
	 */
	public static Product getCatalogue( int number )
	{
		Product theProduct;

		if ( number < AirKM.numCatalogueItems )
		{
			theProduct = catalogue[number];
		}
		else
		{
			System.out.println( "Error: Invalid product number." );
			theProduct = null;
		}
		return theProduct;
	}

	/**
	 * Adds a new transaction to the journal.
	 * <p>
	 * The transaction is created by this method, using the specified
	 * CardHolder, collection of Product items, and transaction category.
	 * 
	 * @param client
	 *            CardHolder associated with the new transaction.
	 * @param transactionCategory
	 *            Type of transaction.
	 * @param items
	 *            Products associated with the transaction.
	 * @return True if the transaction was created successfully, and there is
	 *         space in the journal for the transaction. False otherwise.
	 */
	public static boolean addTransaction( CardHolder client,
			int transactionCategory, Product[] items )
	{
		Transaction newTransaction;
		boolean result;

		if ( AirKM.numJournalItems < AirKM.maxSize )
		{
			newTransaction = new Transaction( client );

			// Determine if the transaction is a purchase or points redemption.

			if ( transactionCategory == Transaction.PURCHASE )
			{
				result = newTransaction.purchase( items );
			}
			else
			{
				result = newTransaction.redeemPoints( items );
			}

			// If creating the transaction was successful, add it
			// to the journal.

			if ( result )
			{
				AirKM.journal[AirKM.numJournalItems] = newTransaction;

				// Increment number of transactions in journal.

				AirKM.numJournalItems = AirKM.numJournalItems + 1;
			}
			else
			{
				// Error message would already have been printed in
				// the purchase() or redeemPoints() method.

				; // Do nothing
			}
		}
		else
		{
			System.out.println( "Error:  Transaction journal is full." );
			result = false;
		}
		return result;
	}

	/**
	 * Create a transaction with only one product, and add the transaction to
	 * the journal.
	 * <p>
	 * This method creates a new Product array with one element, and puts
	 * the product in the array.  Then, the method calls addTransaction()
	 * with the array.
	 * 
	 * @param client CardHolder associated with the transaction.
	 * @param transactionCategory Type of transaction.
	 * @param aProduct Product associated with the transaction.
	 * @return True if the transaction was created successfully, and there is
	 *         space in the journal for the transaction. False otherwise.
	 */
	public static boolean addSingleProductTransaction( CardHolder client,
			int transactionCategory, Product aProduct )
	{
		Product[] products;
		boolean result;

		products = new Product[1];
		products[0] = aProduct;
		result = addTransaction( client, transactionCategory, products );

		return result;
	}

	/**
	 * Prints all transactions associated with a specified card holder identified
	 * by name.
	 * 
	 * @param name Name of the card holder for which to print transactions.
	 */
	public static void printTransactionsForCardHolder( String name )
	{
		int index;
		Transaction aTransaction; 

		System.out.println( "Transactions for " + name );
		System.out.println( "===============================================" );

		// Check each transaction in journal to see if the card holder's name
		// is the one we are looking for.
		
		for ( index = 0; index < AirKM.numJournalItems; index = index + 1 )
		{
			aTransaction = journal[index];
			
			// If the name matches, print the transaction.
			
			if ( aTransaction.getCardHolderName( ).compareTo( name ) == 0 )
			{
				System.out.println( aTransaction );
			}
			else
			{
				; // Do nothing
			}
		}
		System.out.println( );
	}

	/**
	 * Prints a summary of the transacdtion in the journal.
	 * <p>
	 * The values printed are:
	 * <ul>
	 * <li>Number of purchase transactions, with total paid and total points awarded.
	 * <li>Number of redemption transaction, with total points used.
	 * </ul>
	 */
	public static void printSummary( )
	{
		int index; 
		Transaction aTransaction;
		int purchases;	// number of purchase transactions
		int redemptions; // number of redemption transactions
		int pointsAwarded;
		int pointsUsed;
		double totalPaid;
		
		// Initialize summary totals.
		
		purchases = 0;
		redemptions = 0;
		totalPaid = 0.0;
		pointsAwarded = 0;
		pointsUsed = 0;
		
		// Collect information from every transaction in the journal.
		
		for ( index = 0; index < AirKM.numJournalItems; index = index + 1 )
		{
			// Get the next transaction.
			
			aTransaction = journal[index];
			
			// Determine the transaction category.
			
			if ( aTransaction.getCategory( ) == Transaction.PURCHASE )
			{
				// Add to purchase totals.
				
				purchases = purchases + 1;
				totalPaid = totalPaid + aTransaction.getAmountPaid( );
				pointsAwarded = pointsAwarded + aTransaction.getPoints( );
			}
			else
			{
				if ( aTransaction.getCategory( ) == Transaction.REDEMPTION )
				{
					// Add to redemption totals.
					
					redemptions = redemptions + 1;
					pointsUsed = pointsUsed + aTransaction.getPoints( );
				}
				else
				{
					// Skip incomplete tranactions.
					
					; // Do nothing
				}
			}
		}

		// Print summary totals.
		
		System.out.println( "Summary" );
		System.out.println( "=======" );
		System.out.println( purchases + " purchases ($" + totalPaid + " and "
				+ pointsAwarded + " points)" );
		System.out.println( redemptions + " redemptions (" + pointsUsed
				+ " points)" );
		System.out.println( "=======" );
		System.out.println( );
	}
}
