/**
 * @author Alan Williams
 */
public class Transaction
{
	// CLASS VARIABLES

	/**
	 * Category code constant for an incomplete transaction.
	 */
	public static final int INCOMPLETE = 0;

	/**
	 * Category code constant for a purchase transaction.
	 */
	public static final int PURCHASE = 1;

	/**
	 * Category code constant for a points redemption transaction.
	 */
	public static final int REDEMPTION = 2;

	/**
	 * The next transaction number to be assigned on creation of a new
	 * Transaction.
	 */
	private static int nextNumber = 100;

	// INSTANCE VARIABLES

	/**
	 * Transaction category code number: should be one of INCOMPLETE, PURCHASE,
	 * or REDEMPTION.
	 */
	private int category;

	/**
	 * The transaction number for this Transaction object.
	 */
	private int transactionNumber;

	/**
	 * Card holder associated with the transaction.
	 */
	private CardHolder client;

	/**
	 * Collection of products associated with the transaction.
	 */
	private Product[] products;

	/**
	 * Total amount paid for all products purchased (or 0 if this is a points
	 * redemption transaction.
	 */
	private double amountPaid;

	/**
	 * Number of points obtained or used as part of this transaction.
	 */
	private int points;

	// CONSTRUCTOR

	/**
	 * Constructor for a new transaction.
	 * <p>
	 * Creates a new transaction, assigns the next available transaction number,
	 * associates a CardHolder with the transaction, and sets the category to
	 * INCOMPLETE. Other instance variables are set to default values.
	 * 
	 * @param theClient
	 *            The CardHolder associated with this transaction.
	 */
	public Transaction( CardHolder theClient )
	{
		// This transaction is not completely initialized in this method.
		// Set category to INCOMPLETE until another method tells us whether
		// this will be a purchase or points redemption.

		this.category = Transaction.INCOMPLETE;

		// Copy current transaction number to this transaction.

		this.transactionNumber = Transaction.nextNumber;

		// Increment class variable, so next transaction will get
		// the next available transaction number.

		Transaction.nextNumber = Transaction.nextNumber + 1;

		// Associate CardHolder with transaction.

		this.client = theClient;

		// Set other variables to defaults.

		this.products = null;
		this.amountPaid = 0.0;
		this.points = 0;
	}

	// INSTANCE METHODS

	/**
	 * Returns the tranaction category.
	 * 
	 * @return An <code>int</code> with the category code.
	 */
	public int getCategory( )
	{
		return this.category;
	}

	/**
	 * Returns the transaction number.
	 * 
	 * @return An <code>int</code> representing the transaction number.
	 */
	public int getTransactionNumber( )
	{
		return this.transactionNumber;
	}

	/**
	 * Returns the name of the card holder associated with the transaction.
	 * <p>
	 * The CardHolder object is delegated to return the name.
	 * 
	 * @return A string containing the name of the associated card holder.
	 */
	public String getCardHolderName( )
	{
		return this.client.getName( );
	}

	/**
	 * Returns the total number of points earned or redeemed by this
	 * transaction.
	 * 
	 * @return An <code>int</code> with the number of points.
	 */
	public int getPoints( )
	{
		return this.points;
	}

	/**
	 * Returns the total amount paid for all products in the transaction.
	 * <p>
	 * If points are redeemed, this amount would be 0.0. For a purchase
	 * transaction, this amount is the total cost of all associated products.
	 * 
	 * @return A <code>double</code> with the total amount paid during the
	 *         transaction.
	 */
	public double getAmountPaid( )
	{
		return this.amountPaid;
	}

	/**
	 * Complete a purchase transaction of a specified collection of products.
	 * <p>
	 * This method performs a purchase transaction. If the transaction is
	 * already complete, the method returns false. Otherwise, the categeory is
	 * set to PURCHASE.
	 * <p>
	 * A new <code>products</code> array is created, and the collection of
	 * products is copied into that array so that the Transaction has a separate
	 * independent copy of the array. As the copy is being done, the cost of
	 * each product is determined, and added to a running total. The total
	 * amount paid is stored with the transaction.
	 * <p>
	 * Based on the total amount paid, the number of reward points earned is
	 * calculated, stored with the transaction, and also added to the card
	 * holder's total.
	 * 
	 * @param itemList
	 *            Collection of products that are being purchased.
	 * @return True if the transaction is completed successfully, and false if
	 *         this transaction was already a purchase or points redemption.
	 */
	public boolean purchase( Product[] itemList )
	{
		int index;
		boolean success;

		// This method should complete an unfinished transaction.  Otherwise,
		// print an error message.
		
		if ( this.getCategory() == Transaction.INCOMPLETE )
		{
			// Create a copy of the 'itemList' array, and store in 'products'.

			this.products = new Product[itemList.length];
			this.amountPaid = 0.0;
			for ( index = 0; index < itemList.length; index = index + 1 )
			{
				this.products[index] = itemList[index];

				// Find the cost of each item, and add to total.

				this.amountPaid = this.amountPaid + itemList[index].getCost( );
			}

			// Determine number of points earned. Note carefully where the
			// truncation to an integer happens: after the amount paid has
			// been multiplied by 10, but before the multiplication by the
			// bonus factor.

			this.points = ( ( int ) ( this.amountPaid * 10 ) )
					* AirKM.POINTS_FACTOR;

			// Add points to CardHolder's total.

			this.client.adjustPoints( this.getPoints() );

			// Set the transaction category.

			this.category = Transaction.PURCHASE;
			success = true;
		}
		else
		{
			System.out.println( "Error:  Transaction already completed." );
			success = false;
		}
		return success;
	}

	/**
	 * Complete a points redemption transaction for a specified collection of
	 * products.
	 * <p>
	 * This method performs a redemption transaction. If the transaction is
	 * already complete, the method returns false. Otherwise, the categeory is
	 * set to REDEMPTION.
	 * <p>
	 * A new <code>products</code> array is created, and the collection of
	 * products is copied into that array so that the Transaction has a separate
	 * independent copy of the array. As the copy is being done, the number of
	 * points required to obtain each product is determined, and added to a
	 * running total.
	 * <p>
	 * Next, determine if the CardHolder has a sufficient number of points to
	 * get all of the products. If so, store the number of points with the
	 * transaction, and deduct the number of points from the CardHolder. If not,
	 * an error message is printed.
	 * 
	 * @param itemList
	 *            Collection of products that are being obtained from points.
	 * @return True if the transaction is completed successfully, and false if
	 *         this transaction was already a purchase or points redemption, or
	 *         the CardHolder did not have enough points.
	 */
	public boolean redeemPoints( Product[] itemList )
	{
		int index;
		int pointsRequired;
		boolean success;

		// This method should complete an unfinished transaction.  Otherwise,
		// print an error message.
		
		if ( this.getCategory() == Transaction.INCOMPLETE )
		{
			// Determine total number of points required for the transaction.
			
			pointsRequired = 0;
			for ( index = 0; index < itemList.length; index = index + 1 )
			{
				// Get each product's points value.
				
				pointsRequired = pointsRequired
						+ itemList[index].getPointsRequired( );
			}

			// Check to see if the CardHolder has enough points.  If not,
			// print an error message and return false.
			
			if ( !CardHolder.enoughPoints( this.client, pointsRequired ) )
			{
				System.out.print( "Transaction incomplete:  " );
				System.out.println( "Insufficient points to obtain products." );
				success = false;
			}
			else
			{
				// CardHolder has sufficient points.  Copy the collection of
				// products.
				
				this.products = new Product[itemList.length];
				for ( index = 0; index < itemList.length; index = index + 1 )
				{
					this.products[index] = itemList[index];
				}
				
				// Deduct the points used from the CardHolder.
				
				this.client.adjustPoints( -pointsRequired );
				
				// Save other transaction information.
				
				this.points = pointsRequired;
				this.category = Transaction.REDEMPTION;
				success = true;
			}
		}
		else
		{
			System.out.println( "Error:  Transaction is already completed." );
			success = false;
		}
		return success;
	}

	/**
	 * This method converts a category code number to a more user-friendly
	 * message.
	 * 
	 * @param aCategory
	 *            An <code>int</code> representing a category code.
	 * @return A String with the corresponding message.
	 */
	private static String categoryToString( int aCategory )
	{
		String result;

		if ( aCategory == Transaction.PURCHASE )
		{
			result = "Purchase";
		}
		else
		{
			if ( aCategory == Transaction.REDEMPTION )
			{
				result = "Points redemption";
			}
			else
			{
				result = "Incomplete";
			}
		}
		return result;
	}

	/**
	 * Returns a string representation of the transaction.
	 * <p>
	 * 
	 * @return A string that contains the transaction information.
	 */
	@Override
	public String toString( )
	{
		int index;
		String result;

		// The following obtains the new-line character(s) for the system
		// on which this code is running.

		final String newLine = System.getProperty( "line.separator" );

		if ( this.getCategory() == Transaction.INCOMPLETE )
		{
			result = "Transaction incomplete." + newLine;
		}
		else
		{
			result = "Transaction number: " + this.getTransactionNumber( )
					+ newLine;
			result = result + "Card holder: " + this.getCardHolderName( )
					+ newLine;

			// Use categoryToString() to print a more user-friendly message
			// for the category type.

			result = result + "Category: " + categoryToString( this.getCategory() )
					+ newLine;
			result = result + "Products:" + newLine;

			// Get name of each product.

			for ( index = 0; index < this.products.length; index = index + 1 )
			{
				result = result + "  - " + this.products[index].getName( )
						+ newLine;
			}

			if ( this.category == Transaction.PURCHASE )
			{
				result = result + "Amount paid: " + this.getAmountPaid() + newLine;
				result = result + "Points obtained: " + this.getPoints() + newLine;
			}
			else
			{
				result = result + "Points used: " + this.getPoints() + newLine;
			}
		}
		return result;
	}
}